import calendar
import collections
import datetime
import gc
import json
import os
import platform
import time

import statsd
from twisted.application import internet, service
from twisted.internet.defer import succeed
from twisted.internet.error import ReactorAlreadyInstalledError
from twisted.python import log
from twisted.python.log import FileLogObserver, ILogObserver, textFromEventDict, _safeFormat
from twisted.python.logfile import DailyLogFile
from twisted.python.util import untilConcludes
from twisted.web import server
from txjsonrpc_ng.web import jsonrpc
from txjsonrpc_ng.web.data import CacheableResult
from txjsonrpc_ng.web.jsonrpc import with_request

from blitzortung.gis.constants import grid, global_grid
from blitzortung.gis.local_grid import LocalGrid
from blitzortung.service.cache import ServiceCache
from blitzortung.util import TimeConstraint

try:
    from twisted.internet import epollreactor as reactor, defer
except ImportError:
    from twisted.internet import kqreactor as reactor

try:
    reactor.install()
except ReactorAlreadyInstalledError:
    pass

import blitzortung.cache
import blitzortung.config
import blitzortung.db
import blitzortung.geom
import blitzortung.service
from blitzortung.db.query import TimeInterval
from blitzortung.service.db import create_connection_pool
from blitzortung.service.general import create_time_interval
from blitzortung.service.strike_grid import GridParameters

is_pypy = platform.python_implementation() == 'PyPy'

statsd_client = statsd.StatsClient('localhost', 8125, prefix='org.blitzortung.service')

FORBIDDEN_IPS = {}

USER_AGENT_PREFIX = 'bo-android-'


class Blitzortung(jsonrpc.JSONRPC):
    """
    Blitzortung.org JSON-RPC webservice for lightning strike data.

    Provides endpoints for querying strike data, grid-based visualizations,
    and histograms with caching and rate limiting.
    """

    # Grid validation constants
    MIN_GRID_BASE_LENGTH = 5000
    INVALID_GRID_BASE_LENGTH = 1000001
    GLOBAL_MIN_GRID_BASE_LENGTH = 10000

    # Time validation constants
    MAX_MINUTES_PER_DAY = 24 * 60  # 1440 minutes
    DEFAULT_MINUTE_LENGTH = 60
    HISTOGRAM_MINUTE_THRESHOLD = 10

    # User agent validation constants
    MAX_COMPATIBLE_ANDROID_VERSION = 177

    # Memory info interval
    MEMORY_INFO_INTERVAL = 300  # 5 minutes

    def __init__(self, db_connection_pool, log_directory):
        super().__init__()
        self.connection_pool = db_connection_pool
        self.log_directory = log_directory
        self.strike_query = blitzortung.service.strike_query()
        self.strike_grid_query = blitzortung.service.strike_grid_query()
        self.global_strike_grid_query = blitzortung.service.global_strike_grid_query()
        self.histogram_query = blitzortung.service.histogram_query()
        self.check_count = 0
        self.cache = ServiceCache()
        self.current_period = self.__current_period()
        self.current_data = collections.defaultdict(list)
        self.next_memory_info = 0.0
        self.minute_constraints = TimeConstraint(self.DEFAULT_MINUTE_LENGTH, self.MAX_MINUTES_PER_DAY)

    addSlash = True

    def __get_epoch(self, timestamp):
        return calendar.timegm(timestamp.timetuple()) * 1000000 + timestamp.microsecond

    def __current_period(self):
        return datetime.datetime.now(datetime.UTC).replace(second=0, microsecond=0)

    def __check_period(self):
        if self.current_period != self.__current_period():
            self.current_data['timestamp'] = self.__get_epoch(self.current_period)
            if log_directory:
                with open(os.path.join(log_directory, self.current_period.strftime("%Y%m%d-%H%M.json")),
                          'w') as output_file:
                    output_file.write(json.dumps(self.current_data))
            self.__restart_period()

    def __restart_period(self):
        self.current_period = self.__current_period()
        self.current_data = collections.defaultdict(list)

    def __force_range(self, number, min_number, max_number):
        if number < min_number:
            return min_number
        elif number > max_number:
            return max_number
        else:
            return number
            # return max(min_number, min(max_number, number))

    def jsonrpc_check(self):
        self.check_count += 1
        return {'count': self.check_count}

    @with_request
    def jsonrpc_get_strikes(self, request, minute_length, id_or_offset=0):
        """This endpoint is currently blocked for all requests."""
        minute_length = self.__force_range(minute_length, 0, self.MAX_MINUTES_PER_DAY)

        client = self.get_request_client(request)
        user_agent = request.getHeader("User-Agent")
        log.msg('get_strikes(%d, %d) %s %s BLOCKED' % (minute_length, id_or_offset, client, user_agent))
        return None

    def get_strikes_grid(self, minute_length, grid_baselength, minute_offset, region, count_threshold):
        grid_parameters = GridParameters(grid[region].get_for(grid_baselength), grid_baselength, region,
                                         count_threshold=count_threshold)
        time_interval = create_time_interval(minute_length, minute_offset)

        grid_result, state = self.strike_grid_query.create(grid_parameters, time_interval, self.connection_pool,
                                                           statsd_client)

        histogram_result = self.get_histogram(time_interval, envelope=grid_parameters.grid) \
            if minute_length > self.HISTOGRAM_MINUTE_THRESHOLD else succeed([])

        combined_result = self.strike_grid_query.combine_result(grid_result, histogram_result, state)

        combined_result.addCallback(lambda value: CacheableResult(value))

        return combined_result

    def get_global_strikes_grid(self, minute_length, grid_baselength, minute_offset, count_threshold):
        grid_parameters = GridParameters(global_grid.get_for(grid_baselength), grid_baselength,
                                         count_threshold=count_threshold)
        time_interval = create_time_interval(minute_length, minute_offset)

        grid_result, state = self.global_strike_grid_query.create(grid_parameters, time_interval, self.connection_pool,
                                                                  statsd_client)

        histogram_result = self.get_histogram(
            time_interval) if minute_length > self.HISTOGRAM_MINUTE_THRESHOLD else succeed([])

        combined_result = self.strike_grid_query.combine_result(grid_result, histogram_result, state)

        combined_result.addCallback(lambda value: CacheableResult(value))

        return combined_result

    def get_local_strikes_grid(self, x, y, grid_baselength, minute_length, minute_offset, count_threshold, data_area=5):
        local_grid = LocalGrid(data_area=data_area, x=x, y=y)
        grid_factory = local_grid.get_grid_factory()
        grid_parameters = GridParameters(grid_factory.get_for(grid_baselength), grid_baselength,
                                         count_threshold=count_threshold)
        time_interval = create_time_interval(minute_length, minute_offset)

        grid_result, state = self.strike_grid_query.create(grid_parameters, time_interval, self.connection_pool,
                                                           statsd_client)

        histogram_result = self.get_histogram(time_interval, envelope=grid_parameters.grid) \
            if minute_length > self.HISTOGRAM_MINUTE_THRESHOLD else succeed([])

        combined_result = self.strike_grid_query.combine_result(grid_result, histogram_result, state)

        combined_result.addCallback(lambda value: CacheableResult(value))

        return combined_result

    @with_request
    def jsonrpc_get_strikes_raster(self, request, minute_length, grid_base_length=10000, minute_offset=0, region=1):
        return self.jsonrpc_get_strikes_grid(request, minute_length, grid_base_length, minute_offset, region)

    @with_request
    def jsonrpc_get_strokes_raster(self, request, minute_length, grid_base_length=10000, minute_offset=0, region=1):
        return self.jsonrpc_get_strikes_grid(request, minute_length, grid_base_length, minute_offset, region)

    @with_request
    def jsonrpc_get_global_strikes_grid(self, request, minute_length, grid_base_length=10000, minute_offset=0,
                                        count_threshold=0):
        self.memory_info()
        client = self.get_request_client(request)
        user_agent, user_agent_version = self.parse_user_agent(request)

        if client in FORBIDDEN_IPS or user_agent_version == 0 or request.getHeader(
                'content-type') != 'text/json' or request.getHeader(
            'referer') == '' or grid_base_length < self.MIN_GRID_BASE_LENGTH or grid_base_length == self.INVALID_GRID_BASE_LENGTH:
            log.msg(
                f"FORBIDDEN - client: {client}, user agent: {user_agent_version}, content type: {request.getHeader('content-type')}, referer: {request.getHeader('referer')}")
            log.msg('get_global_strikes_grid(%d, %d, %d, >=%d) BLOCKED %.1f%% %s %s' % (
                minute_length, grid_base_length, minute_offset, count_threshold,
                0, client, user_agent))
            return {}

        original_grid_base_length = grid_base_length
        grid_base_length = max(self.GLOBAL_MIN_GRID_BASE_LENGTH, grid_base_length)
        minute_length, minute_offset = self.minute_constraints.enforce(minute_length, minute_offset, )
        count_threshold = max(0, count_threshold)

        cache = self.cache.global_strikes(minute_offset)
        response = cache.get(self.get_global_strikes_grid, minute_length=minute_length,
                             grid_baselength=grid_base_length,
                             minute_offset=minute_offset,
                             count_threshold=count_threshold)
        self.fix_bad_accept_header(request, user_agent)

        log.msg('get_global_strikes_grid(%d, %d, %d, >=%d) %.1f%% %s %s' % (
            minute_length, grid_base_length, minute_offset, count_threshold,
            cache.get_ratio() * 100, client, user_agent))

        self.__check_period()
        self.current_data['get_strikes_grid'].append(
            (self.__get_epoch(datetime.datetime.now(datetime.UTC)), minute_length, original_grid_base_length,
             minute_offset,
             0, count_threshold, client, user_agent))

        statsd_client.incr('strikes_grid.total_count')
        statsd_client.incr('global_strikes_grid.total_count')
        statsd_client.gauge('global_strikes_grid.cache_hits', cache.get_ratio())
        if minute_length == 10:
            statsd_client.incr('strikes_grid.bg_count')
            statsd_client.incr('global_strikes_grid.bg_count')

        return response

    @with_request
    def jsonrpc_get_local_strikes_grid(self, request, x, y, grid_base_length=10000, minute_length=60, minute_offset=0,
                                       count_threshold=0, data_area=5):
        self.memory_info()
        client = self.get_request_client(request)
        user_agent, user_agent_version = self.parse_user_agent(request)

        if client in FORBIDDEN_IPS or request.getHeader(
                'content-type') != 'text/json' or request.getHeader(
            'referer') == '' or grid_base_length < self.MIN_GRID_BASE_LENGTH or grid_base_length == self.INVALID_GRID_BASE_LENGTH:
            log.msg(
                f"FORBIDDEN - client: {client}, user agent: {user_agent_version}, content type: {request.getHeader('content-type')}, referer: {request.getHeader('referer')}")
            log.msg('get_local_strikes_grid(%d, %d, %d, %d, %d, >=%d, %d) BLOCKED %.1f%% %s %s' % (
                x, y, grid_base_length, minute_length, minute_offset, count_threshold, data_area,
                0, client, user_agent))
            return {}

        original_grid_base_length = grid_base_length
        grid_base_length = max(self.MIN_GRID_BASE_LENGTH, grid_base_length)
        minute_length, minute_offset = self.minute_constraints.enforce(minute_length, minute_offset, )
        count_threshold = max(0, count_threshold)
        data_area = round(max(5, data_area))

        cache = self.cache.local_strikes(minute_offset)
        response = cache.get(self.get_local_strikes_grid, x=x, y=y,
                             grid_baselength=grid_base_length,
                             minute_length=minute_length,
                             minute_offset=minute_offset,
                             count_threshold=count_threshold,
                             data_area=data_area)

        log.msg('get_local_strikes_grid(%d, %d, %d, %d, %d, >=%d, %d) %.1f%% %d# %s %s' % (
            x, y, minute_length, grid_base_length, minute_offset, count_threshold, data_area,
            cache.get_ratio() * 100, cache.get_size(), client,
            user_agent))

        self.__check_period()
        self.current_data['get_strikes_grid'].append(
            (
                self.__get_epoch(datetime.datetime.now(datetime.UTC)), minute_length, original_grid_base_length,
                minute_offset,
                -1, count_threshold, client, user_agent, x, y, data_area))

        statsd_client.incr('strikes_grid.total_count')
        statsd_client.incr('local_strikes_grid.total_count')
        statsd_client.incr(f'local_strikes_grid.data_area.{data_area}')
        statsd_client.gauge('local_strikes_grid.cache_hits', cache.get_ratio())
        if minute_length == 10:
            statsd_client.incr('strikes_grid.bg_count')
            statsd_client.incr('local_strikes_grid.bg_count')

        return response

    @with_request
    def jsonrpc_get_strikes_grid(self, request, minute_length, grid_base_length=10000, minute_offset=0, region=1,
                                 count_threshold=0):
        self.memory_info()
        client = self.get_request_client(request)
        user_agent, user_agent_version = self.parse_user_agent(request)

        if client in FORBIDDEN_IPS or user_agent_version == 0 or request.getHeader(
                'content-type') != 'text/json' or request.getHeader(
            'referer') == '' or grid_base_length < self.MIN_GRID_BASE_LENGTH or grid_base_length == self.INVALID_GRID_BASE_LENGTH:
            log.msg(
                f"FORBIDDEN - client: {client}, user agent: {user_agent_version}, content type: {request.getHeader('content-type')}, referer: {request.getHeader('referer')}")
            log.msg('get_strikes_grid(%d, %d, %d, %d, >=%d) BLOCKED %.1f%% %s %s' % (
                minute_length, grid_base_length, minute_offset, region, count_threshold,
                0, client, user_agent))
            return {}

        original_grid_base_length = grid_base_length
        grid_base_length = max(self.MIN_GRID_BASE_LENGTH, grid_base_length)
        minute_length, minute_offset = self.minute_constraints.enforce(minute_length, minute_offset, )
        region = max(1, region)
        count_threshold = max(0, count_threshold)

        cache = self.cache.strikes(minute_offset)
        response = cache.get(self.get_strikes_grid, minute_length=minute_length,
                             grid_baselength=grid_base_length,
                             minute_offset=minute_offset, region=region,
                             count_threshold=count_threshold)
        self.fix_bad_accept_header(request, user_agent)

        log.msg('get_strikes_grid(%d, %d, %d, %d, >=%d) %.1f%% %s %s' % (
            minute_length, grid_base_length, minute_offset, region, count_threshold,
            cache.get_ratio() * 100, client, user_agent))

        self.__check_period()
        self.current_data['get_strikes_grid'].append(
            (self.__get_epoch(datetime.datetime.now(datetime.UTC)), minute_length, original_grid_base_length,
             minute_offset,
             region,
             count_threshold, client, user_agent))

        statsd_client.incr('strikes_grid.total_count')
        statsd_client.incr('strikes_grid.total_count.{}'.format(region))
        statsd_client.gauge('strikes_grid.cache_hits', cache.get_ratio())
        if minute_length == 10:
            statsd_client.incr('strikes_grid.bg_count')
            statsd_client.incr('strikes_grid.bg_count.{}'.format(region))

        return response

    def parse_user_agent(self, request):
        """Parse user agent string to extract version information."""
        user_agent = request.getHeader("User-Agent")
        if user_agent and user_agent.startswith(USER_AGENT_PREFIX):
            user_agent_parts = user_agent.split(' ')[0].rsplit('-', 1)
            version_string = user_agent_parts[1] if len(user_agent_parts) > 1 else None
            user_agent_version = int(version_string) if user_agent_parts[0] == 'bo-android' else 0
        else:
            user_agent_version = 0
        return user_agent, user_agent_version

    def fix_bad_accept_header(self, request, user_agent):
        """Remove Accept-Encoding header for old Android client versions that have bugs."""
        if user_agent and user_agent.startswith(USER_AGENT_PREFIX):
            user_agent_parts = user_agent.split(' ')[0].rsplit('-', 1)
            if len(user_agent_parts) > 1 and user_agent_parts[0] == 'bo-android':
                try:
                    version = int(user_agent_parts[1])
                    if version <= self.MAX_COMPATIBLE_ANDROID_VERSION:
                        request.requestHeaders.removeHeader("Accept-Encoding")
                except ValueError:
                    pass

    def get_histogram(self, time_interval: TimeInterval, region=None, envelope=None):
        return self.cache.histogram.get(self.histogram_query.create,
                                        time_interval=time_interval,
                                        connection_pool=self.connection_pool,
                                        region=region,
                                        envelope=envelope)

    def get_request_client(self, request):
        forward = request.getHeader("X-Forwarded-For")
        if forward:
            return forward.split(', ')[0]
        return request.getClientIP()

    def memory_info(self):
        now = time.time()
        if now > self.next_memory_info:
            log.msg("### MEMORY INFO ###")
            log.msg(gc.get_stats(True) if is_pypy else gc.get_stats())
            self.next_memory_info = now + self.MEMORY_INFO_INTERVAL


class LogObserver(FileLogObserver):

    def __init__(self, f, prefix=None):
        prefix = '' if prefix is None else prefix
        if len(prefix) > 0:
            prefix += ''
        self.prefix = prefix
        FileLogObserver.__init__(self, f)

    def emit(self, event_dict):
        text = textFromEventDict(event_dict)
        if text is None:
            return
        timeStr = self.formatTime(event_dict["time"])
        msgStr = _safeFormat("[%(prefix)s] %(text)s\n", {
            "prefix": self.prefix,
            "text": text.replace("\n", "\n\t")
        })
        untilConcludes(self.write, timeStr + " " + msgStr)
        untilConcludes(self.flush)


application = service.Application("Blitzortung.org JSON-RPC Server")

if os.environ.get('BLITZORTUNG_TEST'):
    import tempfile

    log_directory = tempfile.mkdtemp()
    print("LOG_DIR", log_directory)
else:
    log_directory = "/var/log/blitzortung"
if os.path.exists(log_directory):
    logfile = DailyLogFile("webservice.log", log_directory)
    application.setComponent(ILogObserver, LogObserver(logfile).emit)
else:
    log_directory = None


def start_server(connection_pool):
    print("Connection pool is ready")
    root = Blitzortung(connection_pool, log_directory)
    config = blitzortung.config.config()
    site = server.Site(root)
    site.displayTracebacks = False
    jsonrpc_server = internet.TCPServer(config.get_webservice_port(), site, interface='127.0.0.1')
    jsonrpc_server.setServiceParent(application)
    return jsonrpc_server


def on_error(failure):
    log.err(failure, "Failed to create connection pool")
    raise failure.value


deferred_connection_pool = create_connection_pool()
deferred_connection_pool.addCallback(start_server).addErrback(on_error)



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import datetime
import time

from .data import Timestamp, Timedelta


class Timer():
    """
    simple timer for repeated time measurements
    """

    def __init__(self):
        self.start_time = time.time()
        self.lap_time = self.start_time

    def read(self):
        """
        read time duration of current lap
        """
        return time.time() - self.start_time

    def lap(self):
        """
        start a new lap returning the total time of the currently active lap
        """
        now = time.time()
        lap = now - self.lap_time
        self.lap_time = now
        return lap


def total_seconds(time_value):
    """
    return the total seconds of the given time or datetime (relative to midnight)
    """

    if isinstance(time_value, datetime.datetime) or isinstance(time_value, Timestamp):
        return time_value.hour * 3600 + time_value.minute * 60 + time_value.second
    elif isinstance(time_value, datetime.timedelta) or isinstance(time_value, Timedelta):
        return time_value.seconds + time_value.days * 24 * 3600
    else:
        raise ValueError("unhandled type " + str(type(time_value)))


def round_time(time_value, duration):
    """
    round time to a given timedelta duration
    """
    duration_seconds = total_seconds(duration)
    seconds = (total_seconds(time_value) // duration_seconds) * duration_seconds

    updated_values = {
        'hour': seconds // 3600,
        'minute': seconds // 60 % 60,
        'second': seconds % 60,
        'microsecond': 0
    }

    if isinstance(time_value, Timestamp):
        updated_values['nanosecond'] = 0

    return time_value.replace(**updated_values)


def time_intervals(start_time, duration, end_time=None):
    """
    generator for time interval start times for a specified duration and end time

    if end time is not specified the current time is assumed
    """
    current_time = round_time(start_time, duration)
    if not end_time:
        end_time = datetime.datetime.now(datetime.timezone.utc)
    end_time = round_time(end_time, duration)

    while current_time <= end_time:
        yield current_time
        current_time += duration


def force_range(lower_limit, value, upper_limit):
    if value < lower_limit:
        return lower_limit
    elif value > upper_limit:
        return upper_limit
    else:
        return value

class TimeConstraint:

    def __init__(self, default_minute_length: int, max_minute_length: int):
        self.default_minute_length = default_minute_length
        self.max_minute_length = max_minute_length

    def enforce(self, minute_length, minute_offset):
        minute_length = force_range(0, minute_length, self.max_minute_length)
        minute_length = self.default_minute_length if minute_length == 0 else minute_length
        minute_offset = force_range(-self.max_minute_length + minute_length, minute_offset, 0)
        return minute_length, minute_offset



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from . import histogram, strike, strike_grid
from .histogram import HistogramQuery
from .strike import StrikeQuery
from .strike_grid import GlobalStrikeGridQuery, StrikeGridQuery


def strike_query() -> StrikeQuery:
    from .. import INJECTOR

    return INJECTOR.get(strike.StrikeQuery)


def strike_grid_query() -> StrikeGridQuery:
    from .. import INJECTOR

    return INJECTOR.get(strike_grid.StrikeGridQuery)


def global_strike_grid_query() -> GlobalStrikeGridQuery:
    from .. import INJECTOR

    return INJECTOR.get(strike_grid.GlobalStrikeGridQuery)


def histogram_query() -> HistogramQuery:
    from .. import INJECTOR

    return INJECTOR.get(histogram.HistogramQuery)



from ..cache import ObjectCache


class ServiceCache:
    CACHE_CLEANUP_PERIOD = 300  # 5 minutes
    CACHE_TTL_SHORT = 20  # seconds
    CACHE_TTL_LONG = 60  # seconds
    LOCAL_CACHE_SIZE_CURRENT = 100
    LOCAL_CACHE_SIZE_HISTORY = 400

    def __init__(self):
        self.__strikes_grid = ObjectCache(
            ttl_seconds=self.CACHE_TTL_SHORT, cleanup_period=self.CACHE_CLEANUP_PERIOD)
        self.__strikes_history_grid = ObjectCache(
            ttl_seconds=self.CACHE_TTL_LONG, cleanup_period=self.CACHE_CLEANUP_PERIOD)

        self.__global_strikes_grid = ObjectCache(
            ttl_seconds=self.CACHE_TTL_SHORT, cleanup_period=self.CACHE_CLEANUP_PERIOD)
        self.__global_strikes_history_grid = ObjectCache(
            ttl_seconds=self.CACHE_TTL_LONG, cleanup_period=self.CACHE_CLEANUP_PERIOD)

        self.__local_strikes_grid = ObjectCache(
            ttl_seconds=self.CACHE_TTL_SHORT, size=self.LOCAL_CACHE_SIZE_CURRENT,
            cleanup_period=self.CACHE_CLEANUP_PERIOD)
        self.__local_strikes_history_grid = ObjectCache(
            ttl_seconds=self.CACHE_TTL_LONG, size=self.LOCAL_CACHE_SIZE_HISTORY,
            cleanup_period=self.CACHE_CLEANUP_PERIOD)

        self.histogram = ObjectCache(
            ttl_seconds=self.CACHE_TTL_LONG, cleanup_period=self.CACHE_CLEANUP_PERIOD)

    def global_strikes(self, minute_offset):
        return self.__global_strikes_grid if minute_offset == 0 else self.__global_strikes_history_grid

    def local_strikes(self, minute_offset):
        return self.__local_strikes_grid if minute_offset == 0 else self.__local_strikes_history_grid

    def strikes(self, minute_offset):
        return self.__strikes_grid if minute_offset == 0 else self.__strikes_history_grid



"""

   Copyright 2025 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import psycopg2
import psycopg2.extras
from twisted.internet.defer import Deferred
from twisted.python import log
from txpostgres import reconnection
from txpostgres.txpostgres import Connection, ConnectionPool

import blitzortung.config
from blitzortung.db.query import SelectQuery


def connection_factory(*args, **kwargs):
    """Create a psycopg2 connection with DictConnection factory."""
    kwargs['connection_factory'] = psycopg2.extras.DictConnection
    return psycopg2.connect(*args, **kwargs)


class LoggingDetector(reconnection.DeadConnectionDetector):
    """Database connection detector that logs reconnection events."""

    def startReconnecting(self, f):
        print('[*] database connection is down (error: %r)' % f.value)
        return reconnection.DeadConnectionDetector.startReconnecting(self, f)

    def reconnect(self):
        print('[*] reconnecting...')
        return reconnection.DeadConnectionDetector.reconnect(self)

    def connectionRecovered(self):
        print('[*] connection recovered')
        return reconnection.DeadConnectionDetector.connectionRecovered(self)


class DictConnection(Connection):
    """Database connection using DictConnection factory with logging detector."""
    connectionFactory = staticmethod(connection_factory)

    def __init__(self, reactor=None, cooperator=None, detector=None):
        if not detector:
            detector = LoggingDetector()
        super(DictConnection, self).__init__(reactor, cooperator, detector)


class DictConnectionPool(ConnectionPool):
    """Connection pool using DictConnection instances."""
    connectionFactory = DictConnection

    def __init__(self, _ignored, *connargs, **connkw):
        super(DictConnectionPool, self).__init__(_ignored, *connargs, **connkw)


def create_connection_pool() -> Deferred:
    """Create and start the database connection pool."""
    config = blitzortung.config.config()
    db_connection_string = config.get_db_connection_string()

    connection_pool = DictConnectionPool(None, db_connection_string)

    d = connection_pool.start()
    d.addErrback(log.err)

    return d


def execute(connection, query: SelectQuery):
    return connection.runQuery(str(query), query.get_parameters())



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import atexit

from injector import Module, singleton, inject, provider

from . import compat  # Register psycopg2cffi compatibility

import psycopg2
import psycopg2.pool
import psycopg2.extras
import psycopg2.extensions

from .. import config

from . import query, query_builder, mapper, table


class DbModule(Module):
    @staticmethod
    def cleanup(connection_pool):
        connection_pool.closeall()

    @singleton
    @provider
    @inject
    def provide_psycopg2_connection_pool(self, config: config.Config) -> psycopg2.pool.ThreadedConnectionPool:
        connection_pool = psycopg2.pool.ThreadedConnectionPool(4, 50, config.get_db_connection_string())
        atexit.register(self.cleanup, connection_pool)
        return connection_pool


def strike():
    from .. import INJECTOR

    strike_db = INJECTOR.get(table.Strike)
    return strike_db


def strike_cluster():
    from .. import INJECTOR

    return INJECTOR.get(table.StrikeCluster)


def station():
    from .. import INJECTOR

    return INJECTOR.get(table.Station)


def station_offline():
    from .. import INJECTOR

    return INJECTOR.get(table.StationOffline)


def location():
    from .. import INJECTOR

    return INJECTOR.get(table.Location)


def servicelog_total():
    from .. import INJECTOR

    return INJECTOR.get(table.ServiceLogTotal)


def servicelog_country():
    from .. import INJECTOR

    return INJECTOR.get(table.ServiceLogCountry)


def servicelog_version():
    from .. import INJECTOR

    return INJECTOR.get(table.ServiceLogVersion)


def servicelog_parameters():
    from .. import INJECTOR

    return INJECTOR.get(table.ServiceLogParameters)



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

"""
Database compatibility layer for psycopg2cffi.

This module registers psycopg2cffi as a drop-in replacement for psycopg2,
which is useful for PyPy compatibility. The registration happens automatically
when this module is imported.
"""

try:
    from psycopg2cffi import compat
    compat.register()
except ImportError:
    pass



# -*- coding: utf8 -*-

"""

   Copyright 2014-2022 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import datetime
import math

from . import base
from .geom import GridElement


class Timestamp(base.EqualityAndHash):
    timestamp_string_minimal_fractional_seconds_length = 20
    timestamp_string_microseconds_length = 26

    __slots__ = ['datetime', 'nanosecond']

    def __init__(self, date_time=datetime.datetime.now(datetime.timezone.utc), nanosecond=0):
        if isinstance(date_time, str):
            date_time, date_time_nanosecond = Timestamp.from_timestamp(date_time)
            nanosecond += date_time_nanosecond
        elif isinstance(date_time, int):
            date_time, date_time_nanosecond = Timestamp.from_nanoseconds(date_time)
            nanosecond += date_time_nanosecond

        if nanosecond < 0 or nanosecond > 999:
            microdelta = nanosecond // 1000
            date_time += datetime.timedelta(microseconds=microdelta)
            nanosecond -= microdelta * 1000

        self.datetime = date_time
        self.nanosecond = nanosecond

    @staticmethod
    def from_timestamp(timestamp_string):
        try:
            if len(timestamp_string) > Timestamp.timestamp_string_minimal_fractional_seconds_length:
                divider_index = Timestamp.timestamp_string_microseconds_length
                date_time = datetime.datetime.strptime(timestamp_string[:divider_index], '%Y-%m-%d %H:%M:%S.%f')
                nanosecond_string = timestamp_string[divider_index:]
                nanosecond = int(
                    float(nanosecond_string) * math.pow(10, 3 - len(nanosecond_string))) if nanosecond_string else 0
            else:
                date_time = datetime.datetime.strptime(timestamp_string, '%Y-%m-%d %H:%M:%S')
                nanosecond = 0

            return date_time.replace(tzinfo=datetime.timezone.utc), nanosecond
        except ValueError:
            return None, 0

    @staticmethod
    def from_nanoseconds(total_nanoseconds):
        total_microseconds = total_nanoseconds // 1000
        residual_nanoseconds = total_nanoseconds % 1000
        total_seconds = total_microseconds // 1000000
        residual_microseconds = total_microseconds % 1000000
        return datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc) + \
               datetime.timedelta(seconds=total_seconds,
                                  microseconds=residual_microseconds), \
            residual_nanoseconds

    @property
    def year(self):
        return self.datetime.year

    @property
    def month(self):
        return self.datetime.month

    @property
    def day(self):
        return self.datetime.day

    @property
    def hour(self):
        return self.datetime.hour

    @property
    def minute(self):
        return self.datetime.minute

    @property
    def second(self):
        return self.datetime.second

    @property
    def microsecond(self):
        return self.datetime.microsecond

    @property
    def tzinfo(self):
        return self.datetime.tzinfo

    epoch = datetime.datetime.fromtimestamp(0, datetime.timezone.utc)

    @property
    def value(self):
        total_microseconds = int((self.datetime - self.epoch).total_seconds() * 1000000)
        return (total_microseconds * 1000 + self.nanosecond) if self.datetime is not None else -1

    @property
    def is_valid(self):
        return self.datetime is not None and self.datetime.year > 1900

    def __ne__(self, other):
        return self.datetime != other.datetime or self.nanosecond != other.nanosecond

    def __lt__(self, other):
        if isinstance(other, datetime.datetime):
            return self.datetime < other
        else:
            return self.datetime < other.datetime or (
                    self.datetime == other.datetime and self.nanosecond < other.nanosecond)

    def __le__(self, other):
        if isinstance(other, datetime.datetime):
            return self.datetime <= other
        else:
            return self.datetime < other.datetime or (
                    self.datetime == other.datetime and self.nanosecond <= other.nanosecond)

    def __gt__(self, other):
        if isinstance(other, datetime.datetime):
            return self.datetime > other
        else:
            return self.datetime > other.datetime or (
                    self.datetime == other.datetime and self.nanosecond > other.nanosecond)

    def __ge__(self, other):
        if isinstance(other, datetime.datetime):
            return self.datetime >= other
        else:
            return self.datetime > other.datetime or (
                    self.datetime == other.datetime and self.nanosecond >= other.nanosecond)

    def __add__(self, other):
        if isinstance(other, Timedelta):
            return Timestamp(self.datetime + other.timedelta, self.nanosecond + other.nanodelta)
        elif isinstance(other, datetime.timedelta):
            return Timestamp(self.datetime + other, self.nanosecond)
        elif isinstance(other, int):
            return Timestamp(self.datetime, self.nanosecond + other)
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Timestamp):
            return Timedelta(self.datetime - other.datetime, self.nanosecond - other.nanosecond)
        elif isinstance(other, datetime.timedelta):
            return Timestamp(self.datetime - other, self.nanosecond)
        elif isinstance(other, int):
            return Timestamp(self.datetime, self.nanosecond - other)
        return NotImplemented

    def strftime(self, datetime_format):
        return self.datetime.strftime(datetime_format)

    def replace(self, **kwargs):
        if 'nanosecond' in kwargs:
            nanosecond = kwargs['nanosecond']
            del kwargs['nanosecond']
        else:
            nanosecond = self.nanosecond
        return Timestamp(self.datetime.replace(**kwargs), nanosecond)

    def __repr__(self):
        return "Timestamp({}{:03d})".format(self.datetime.strftime("%Y-%m-%d %H:%M:%S.%f"), self.nanosecond)


NaT = Timestamp(None)


class Timedelta(base.EqualityAndHash):
    def __init__(self, timedelta=datetime.timedelta(), nanodelta=0):
        if nanodelta < 0 or nanodelta > 999:
            microdelta = nanodelta // 1000
            timedelta += datetime.timedelta(microseconds=microdelta)
            nanodelta -= microdelta * 1000
        self.timedelta = timedelta
        self.nanodelta = nanodelta

    @property
    def days(self):
        return self.timedelta.days

    @property
    def seconds(self):
        return self.timedelta.seconds

    def __repr__(self):
        return "Timedelta({}, {})".format(self.timedelta, self.nanodelta)


class Event(base.Point):
    time_format = '%Y-%m-%d %H:%M:%S'
    time_format_fractional_seconds = time_format + '.%f'

    __slots__ = ['__timestamp']

    def __init__(self, timestamp, x_coord_or_point, y_coord=None):
        super().__init__(x_coord_or_point, y_coord)
        self.__timestamp = timestamp

    @property
    def timestamp(self):
        return self.__timestamp

    def difference_to(self, other):
        return other.timestamp - self.timestamp

    def ns_difference_to(self, other):
        value1 = other.timestamp.value
        value2 = self.timestamp.value
        difference = value1 - value2
        return difference

    def has_same_location(self, other):
        return super().__eq__(other)

    @property
    def is_valid(self):
        return (not math.isclose(self.x, 0.0, rel_tol=1e-09, abs_tol=1e-09) or not math.isclose(self.y, 0.0, rel_tol=1e-09)) \
            and -180 <= self.x <= 180 \
            and -90 < self.y < 90 \
            and self.has_valid_timestamp

    @property
    def has_valid_timestamp(self):
        return self.timestamp is not None and self.timestamp.is_valid

    def __lt__(self, other):
        return self.timestamp.value < other.timestamp.value

    def __le__(self, other):
        return self.timestamp.value <= other.timestamp.value

    def __str__(self):
        if self.has_valid_timestamp:
            timestamp_string = self.timestamp.strftime(self.time_format_fractional_seconds + 'XXX')
            timestamp_string = timestamp_string.replace('XXX', "%03d" % self.timestamp.nanosecond)
        else:
            timestamp_string = "NaT"

        return "%s %.4f %.4f" \
            % (timestamp_string, self.x, self.y)

    @property
    def uuid(self):
        return "%s-%05.0f-%05.0f" \
            % (str(self.timestamp().value), self.x * 100, self.y * 100)


class Strike(Event):
    """
    class for strike objects
    """

    __slots__ = ['id', 'altitude', 'amplitude', 'lateral_error', 'station_count', 'stations']

    def __init__(self, strike_id, timestamp, x_coord, y_coord, altitude, amplitude, lateral_error, station_count,
                 stations=None):
        super().__init__(timestamp, x_coord, y_coord)
        self.id = strike_id
        self.altitude = altitude
        self.amplitude = amplitude
        self.lateral_error = lateral_error
        self.station_count = station_count
        self.stations = [] if stations is None else stations

    def has_participant(self, participant):
        """
        returns true if the given participant is contained in the stations list
        """
        return participant in self.stations

    def __str__(self):
        return super().__str__() + " %s %.1f %d %d" % (
            str(self.altitude) if self.altitude is not None else '-',
            self.amplitude if self.amplitude else 0.0,
            self.lateral_error,
            self.station_count if self.station_count else 0
        )


class ChannelWaveform:
    """
    class for raw data waveform channels
    """

    def __init__(self, channel_number, amplifier_version, antenna, gain, values, start, bits, shift, conversion_gap,
                 conversion_time, waveform):
        self.channel_number = channel_number
        self.amplifier_version = amplifier_version
        self.antenna = antenna
        self.gain = gain
        self.values = values
        self.start = start
        self.bits = bits
        self.shift = shift
        self.conversion_gap = conversion_gap
        self.conversion_time = conversion_time
        self.waveform = waveform


class GridData:
    """ class for grid characteristics"""

    def __init__(self, grid, no_data=None):
        self.grid = grid
        self.no_data = no_data if no_data else GridElement(0, None)
        self.data = [[None for _ in range(grid.x_bin_count)] for _ in range(grid.y_bin_count)]

    def set(self, x_index, y_index, value):
        try:
            self.data[y_index][x_index] = value
        except IndexError:
            pass

    def get(self, x_index, y_index):
        return self.data[y_index][x_index]

    def to_arcgrid(self):
        result = 'NCOLS %d\n' % self.grid.x_bin_count
        result += 'NROWS %d\n' % self.grid.y_bin_count
        result += 'XLLCORNER %.4f\n' % self.grid.x_min
        result += 'YLLCORNER %.4f\n' % self.grid.y_min
        result += 'CELLSIZE %.4f\n' % self.grid.x_div
        result += 'NODATA_VALUE %s\n' % str(self.no_data.count)

        result += '\n'.join([' '.join([self.cell_to_multiplicity(cell) for cell in row]) for row in self.data[::-1]])

        return result

    @staticmethod
    def cell_to_multiplicity(current_cell):
        return str(current_cell.count) if current_cell else '0'

    def to_map(self):
        chars = " .-o*O8"

        matrix = self.data[::-1]
        maximum, total = self.max_and_total_entries(matrix)

        if maximum > len(chars):
            divider = float(maximum) / (len(chars) - 1)
        else:
            divider = 1

        result = (self.grid.x_bin_count + 2) * '-' + '\n'
        for row in self.data[::-1]:
            result += "|"
            for cell in row:
                index = self.cell_index(cell, divider)
                result += chars[index]
            result += "|\n"

        result += (self.grid.x_bin_count + 2) * '-' + '\n'
        result += 'total count: %d, max per area: %d' % (total, maximum)
        return result

    @staticmethod
    def cell_index(cell, divider):
        return int(math.floor((cell.count - 1) / divider + 1)) if cell else 0

    @staticmethod
    def max_and_total_entries(matrix):
        maximum = 0
        total = 0
        for row in matrix:
            for cell in row:
                if cell:
                    total += cell.count
                    if maximum < cell.count:
                        maximum = cell.count
        return maximum, total

    def to_reduced_array(self, reference_time):

        reduced_array = []

        for row_index, row in enumerate(self.data[::-1]):
            reduced_array += tuple((column_index, row_index,
                                    int(cell.count),
                                    -(reference_time - cell.timestamp).seconds) for column_index, cell in
                                   enumerate(row) if cell)

        return tuple(reduced_array)



from dataclasses import dataclass

import blitzortung.geom
from .constants import UTM_NORTH, UTM_SOUTH


@dataclass
class LocalGrid:
    data_area: int
    x: int
    y: int

    @property
    def size(self):
        return self.data_area * DATA_AREA_SIZE_FACTOR

    @property
    def reference_longitude(self):
       return (self.x - 1) * self.data_area

    @property
    def reference_latitude(self):
        return (self.y - 1) * self.data_area

    @property
    def center_latitude(self):
        return self.reference_latitude + self.size / 2.0

    @property
    def longitude_extension(self):
        return abs(self.center_latitude) / 15.0


    def get_grid_factory(self) -> blitzortung.geom.GridFactory:
        return blitzortung.geom.GridFactory(
            self.reference_longitude - self.longitude_extension,
            self.reference_longitude + self.size + self.longitude_extension,
            self.reference_latitude,
            self.reference_latitude + self.size,
            UTM_NORTH if self.reference_latitude >= 0 else UTM_SOUTH,
            LOCAL_GRID_UTM_LONGITUDE,
            self.reference_latitude + self.size / 2.0
        )


DATA_AREA_SIZE_FACTOR = 3
LOCAL_GRID_UTM_LONGITUDE = 3



"""

Inserts the generated json reports from the service into a compact line-based file format

"""

from __future__ import print_function

import datetime
import glob
import json
import logging
import os
import sys
from optparse import OptionParser

from blitzortung.convert import value_to_string

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel(logging.INFO)

import geoip2.database


def main():
    reader = geoip2.database.Reader('/var/lib/GeoIP/GeoLite2-City.mmdb')

    parser = OptionParser()

    parser.add_option("--debug", dest="debug", default=False, action="store_true", help="enable debug output")

    (options, args) = parser.parse_args()

    if options.debug:
        logger.setLevel(logging.DEBUG)

    latest_time = None
    logger.debug("latest time %s" % latest_time)

    base_dir = '/var/log/blitzortung'

    json_file_names = glob.glob(os.path.join(base_dir, '*.json'))

    json_file_names.sort()

    for json_file_name in json_file_names:

        with open(json_file_name, 'r') as json_file:

            logger.debug(f"opened file {json_file_name}")

            data = json.load(json_file)

            global_timestamp = datetime.datetime.fromtimestamp(data['timestamp'] / 1000000)

            if 'get_strikes_grid' in data:

                results = []
                for entry in data['get_strikes_grid']:
                    remote_address = entry[6]

                    country_code = None
                    city = None
                    local_x = None
                    local_y = None
                    data_area = None

                    user_agent = entry[7]
                    version = None
                    if user_agent:
                        user_agent_parts = user_agent.split(' ')[0].rsplit('-', 1)
                        version_prefix = user_agent_parts[0]
                        version_string = user_agent_parts[1] if len(user_agent_parts) > 1 else None
                        if version_prefix == 'bo-android':
                            version = int(version_string)

                    try:
                        geo_info = reader.city(remote_address)
                        city = geo_info.city.name
                        country_code = geo_info.country.iso_code
                    except (ValueError, geoip2.errors.AddressNotFoundError):
                        pass

                    remote_address = None

                    timestamp_microseconds = entry[0]
                    minute_length = entry[1]
                    grid_baselength = entry[2]
                    minute_offset = entry[3]
                    region = entry[4]
                    count_threshold = entry[5]
                    if len(entry) > 8:
                        local_x = entry[8]
                        local_y = entry[9]
                        data_area = entry[10]

                    results.append([
                        timestamp_microseconds / 1000000,
                        region,
                        grid_baselength,
                        minute_offset,
                        minute_length,
                        count_threshold,
                        remote_address if remote_address is not None else '-',
                        country_code if country_code is not None else '-',
                        city if city is not None else '-',
                        version,
                        local_x if local_x is not None else '-',
                        local_y if local_y is not None else '-',
                        data_area if data_area is not None else '-',
                    ])

                with open(os.path.join(base_dir, "servicelog_" + global_timestamp.strftime("%Y-%m-%d")), 'a+') as output_file:
                    for result in results:
                        line = "\t".join([value_to_string(value) for value in result])
                        output_file.write(line + "\n")

        os.unlink(json_file_name)

if __name__ == "__main__":
    main()



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import datetime
from abc import abstractmethod

import shapely.wkb
from injector import inject

from .. import builder


class ObjectMapper:
    @abstractmethod
    def create_object(self, result, **kwargs):
        pass

    @staticmethod
    def convert_to_timezone(timestamp, target_timezone=None):
        if timestamp is not None:
            target_timezone = target_timezone if target_timezone is not None else datetime.timezone.utc
            timestamp = timestamp.astimezone(target_timezone)
        #            if target_timezone != datetime.timezone.utc:
        #                timestamp = target_timezone.enfold(timestamp)
        return timestamp


class Strike(ObjectMapper):
    @inject
    def __init__(self, strike_builder: builder.Strike):
        self.strike_builder = strike_builder

    def create_object(self, result, **kwargs):
        timezone = kwargs['timezone'] if 'timezone' in kwargs else datetime.timezone.utc

        self.strike_builder.set_id(result['id'])
        self.strike_builder.set_timestamp(
            self.convert_to_timezone(result['timestamp'], timezone),
            result['nanoseconds'])
        self.strike_builder.set_x(result['x'])
        self.strike_builder.set_y(result['y'])
        self.strike_builder.set_altitude(result['altitude'])
        self.strike_builder.set_amplitude(result['amplitude'])
        self.strike_builder.set_station_count(result['stationcount'])
        self.strike_builder.set_lateral_error(result['error2d'])

        return self.strike_builder.build()



# -*- coding: utf8 -*-

"""

   Copyright 2014-2016 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import time


class CacheEntry:
    def __init__(self, payload, expiry_time):
        self.__payload = payload
        self.__expiry_time = expiry_time
        self.__hit_count = 0

    def is_valid(self, current_time):
        return current_time < self.__expiry_time

    def get_payload(self):
        self.__hit_count += 1
        return self.__payload

    def get_hit_count(self):
        return self.__hit_count

    def __repr__(self):
        valid = "+" if self.is_valid(self.__hit_count) else "-"

        return f"cached<{valid} {self.get_hit_count()}>:{self.__payload}"


class ObjectCache:
    kwargs_separator = object()

    def __init__(self, ttl_seconds=30, size=None, cleanup_period=None):
        self.__ttl_seconds = int(ttl_seconds)
        self.total_count = 0
        self.total_hit_count = 0
        self.size = size

        self.cache = {}
        self.keys = {}
        self.last_cleanup = 0
        self.cleanup_period = cleanup_period

    def get(self, cached_object_creator, *args, **kwargs):
        if self.cleanup_period is not None:
            now = time.time()
            if now > self.last_cleanup + self.cleanup_period:
                before = self.get_size()
                self.clean_expired()
                after = self.get_size()
                print(f"{cached_object_creator.__name__}: cache cleanup {before} -> {after}")
                self.last_cleanup = now

        self.total_count += 1

        cache_key = self.generate_cache_key(cached_object_creator, args, kwargs)

        current_time = int(time.time())

        if cache_key in self.cache:
            entry = self.cache[cache_key]
            if entry.is_valid(current_time):
                if self.size is not None:
                    self.track_usage(cache_key)
                self.total_hit_count += 1
                return entry.get_payload()
        elif self.size is not None and len(self.keys) >= self.size:
            self.remove_oldest_entry()

        expires = current_time + self.__ttl_seconds
        payload = cached_object_creator(*args, **kwargs)

        entry = CacheEntry(payload, expires)
        self.cache[cache_key] = entry
        self.keys[cache_key] = 0

        return entry.get_payload()

    def remove_oldest_entry(self):
        expired_key = next(iter(self.keys))
        del self.keys[expired_key]
        del self.cache[expired_key]

    def track_usage(self, cache_key):
        count = self.keys[cache_key]
        del self.keys[cache_key]
        self.keys[cache_key] = count + 1

    def clear(self):
        self.total_count = 0
        self.total_hit_count = 0
        self.cache.clear()

    def clean_expired(self):
        now = time.time()
        expired_keys = {key for key, entry in self.cache.items() if not entry.is_valid(now)}
        for expired_key in expired_keys:
            del self.keys[expired_key]
            del self.cache[expired_key]

    def get_time_to_live(self):
        return self.__ttl_seconds

    def get_ratio(self):
        if self.total_hit_count == 0:
            return 0.0
        return self.total_hit_count / self.total_count

    def get_size(self):
        return len(self.cache)


    def generate_cache_key(self, cached_object_creator, args, kwargs):
        """
        Generates a cache key based on the cached object creator function, args, and kwargs.

        :param cached_object_creator: The function used to create the object to be cached.
        :param args: Positional arguments to the object creator.
        :param kwargs: Keyword arguments to the object creator.
        :return: A tuple representing the cache key.
        """
        return (cached_object_creator,) + args + (self.kwargs_separator,) + tuple(sorted(kwargs.items()))



# -*- coding: utf8 -*-

"""
   Copyright (C) 2011-2025 Andreas Würl

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import datetime
import logging
import os
import time
from contextlib import nullcontext

import requests
import statsd
import stopit
from optparse import OptionParser

import blitzortung.dataimport
import blitzortung.db
import blitzortung.logger
from blitzortung.data import Timestamp
from blitzortung.lock import LockWithTimeout, FailedToAcquireException

logger = logging.getLogger(os.path.basename(__file__))
blitzortung.set_parent_logger(logger)
blitzortung.add_log_handler(blitzortung.logger.create_console_handler())

statsd_client = statsd.StatsClient('localhost', 8125, prefix='org.blitzortung.import')


def timestamp_is_newer_than(timestamp, latest_time):
    if not latest_time:
        return True
    return timestamp and timestamp > latest_time and timestamp - latest_time != datetime.timedelta()


def import_strikes_for(region, start_time, is_update=False):
    logger.debug("work on region %d", region)
    strike_db = blitzortung.db.strike()
    latest_time = strike_db.get_latest_time(region)
    logger.debug("latest time for region %d: %s", region, latest_time)
    if not latest_time:
        latest_time = start_time

    if is_update:
        start_time = update_start_time()
        if not latest_time or start_time > latest_time:
            latest_time = start_time

    reference_time = time.time()
    strike_source = blitzortung.dataimport.strikes()
    strikes = strike_source.get_strikes_since(latest_time, region=region)
    query_time = time.time()

    strike_group_size = 10000
    strike_count = 0
    global_start_time = start_time = time.time()
    for strike in strikes:
        strike_db.insert(strike, region)

        strike_count += 1
        if strike_count % strike_group_size == 0:
            strike_db.commit()
            logger.info("commit #{} ({:.1f}/s) @{} for region {}".format(
                strike_count, strike_group_size / (time.time() - start_time), strike.timestamp, region))
            start_time = time.time()

    if strike_count > 0:
        strike_db.commit()

    insert_time = time.time()
    stat_name = "strikes.%d" % region
    statsd_client.incr(stat_name)
    statsd_client.gauge(stat_name + ".count", strike_count)
    statsd_client.timing(stat_name + ".get", max(1, int((query_time - reference_time) * 1000)))
    statsd_client.timing(stat_name + ".insert", max(1, int((insert_time - query_time) * 1000)))

    logger.info("imported {} strikes ({:.1f}/s) for region {}".format(
        strike_count, strike_count / (time.time() - global_start_time), region))


def update_start_time() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(minutes=5)


def import_strikes(regions, start_time, no_timeout=False, is_update=False):

    error_count = 0
    for region in regions:
        for retry in range(5):
            try:
                with nullcontext() if no_timeout else stopit.SignalTimeout(300):
                    import_strikes_for(region, start_time, is_update=is_update)
                break
            except (requests.exceptions.ConnectionError, stopit.TimeoutException):
                logger.warning('import failed: retry {} region {}'.format(retry, region))
                error_count += 1
                time.sleep(2)
                continue
    statsd_client.gauge("strikes.error_count", error_count)


def main():
    parser = OptionParser()
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true", help="verbose output")
    parser.add_option("-d", "--debug", dest="debug", action="store_true", help="debug output")
    parser.add_option("--no-timeout", dest="no_timeout", action="store_true", help="do not apply 5 minute timeout")
    parser.add_option("--startdate", dest="startdate", default=None, help="import start date")
    parser.add_option("--update", dest="update", action="store_true", help="run as regular update")

    (options, args) = parser.parse_args()

    lock = LockWithTimeout('/tmp/.bo-import.lock')

    try:
        with lock.locked(10):
            if options.debug:
                blitzortung.set_log_level(logging.DEBUG)
            elif options.verbose:
                blitzortung.set_log_level(logging.INFO)
            start_time = Timestamp(datetime.datetime.strptime(options.startdate, "%Y%m%d").replace(
                tzinfo=datetime.timezone.utc)) if options.startdate else None

            regions = range(20)
            import_strikes(regions=regions, start_time=start_time, no_timeout=options.no_timeout, is_update=options.update)
    except FailedToAcquireException:
        logger.warning("could not acquire lock")


if __name__ == "__main__":
    main()



import pyproj

import blitzortung.geom

UTM_EU = pyproj.CRS('epsg:32633')  # UTM 33 N / WGS84
UTM_NORTH_AMERICA = pyproj.CRS('epsg:32614')  # UTM 14 N / WGS84
UTM_CENTRAL_AMERICA = pyproj.CRS('epsg:32614')  # UTM 14 N / WGS84
UTM_SOUTH_AMERICA = pyproj.CRS('epsg:32720')  # UTM 20 S / WGS84
UTM_OCEANIA = pyproj.CRS('epsg:32755')  # UTM 55 S / WGS84
UTM_ASIA = pyproj.CRS('epsg:32650')  # UTM 50 N / WGS84
UTM_AFRICA = pyproj.CRS('epsg:32633')  # UTM 33 N / WGS84
UTM_NORTH = pyproj.CRS('epsg:32631')  # UTM 31 N / WGS84
UTM_SOUTH = pyproj.CRS('epsg:32731')  # UTM 31 S / WGS84

grid = {
    1: blitzortung.geom.GridFactory(-25, 57, 27, 72, UTM_EU),
    2: blitzortung.geom.GridFactory(110, 180, -50, 0, UTM_OCEANIA),
    3: blitzortung.geom.GridFactory(-140, -50, 10, 60, UTM_NORTH_AMERICA),
    4: blitzortung.geom.GridFactory(85, 150, -10, 60, UTM_ASIA),
    5: blitzortung.geom.GridFactory(-100, -30, -50, 20, UTM_SOUTH_AMERICA),
    6: blitzortung.geom.GridFactory(-20, 50, -40, 40, UTM_AFRICA),
    7: blitzortung.geom.GridFactory(-115, -50, 0, 30, UTM_CENTRAL_AMERICA)
}

global_grid = blitzortung.geom.GridFactory(-180, 180, -90, 90, UTM_EU, 11, 48)
